/*

a:

Programmet fastnar:
- Detta kan exempelvis ske om trådbyte sker i samband med att 'threads_done' ökas. Om
  två trådar läser det gamla värdet, ökar det, och skriver tillbaka resultatet så kan
  en uppdatering "tappas bort", vilket gör så att tråden som kör 'fractal_print' aldrig
  kommer att komma förbi loopen där den väntar på att 'threads_done' blir lika med
  'THREAD_COUNT'.

Programmet har inte uppdaterat en eller flera pixlar:
- Detta kan ske på grund av ett liknande problem som ovan, men för 'next_col':
  Exempelvis kan ökningen av 'next_col' implementeras som följande av hårdvara och/eller kompilator:

  int tmp = d->next_col;
  d->next_col++;

  Om två trådar kör koden samtidigt när 'next_col' är 0, så kan båda först läsa det gamla värdet av
  'next_col' till 'tmp', och därmed båda arbeta på kolumn 0. Om de sedan ökar 'next_col' på ett
  korrekt sätt, så kommer 'next_col' att bli 2, utan att någon tråd har behandlat kolumn nummer 1.

b:

Busy wait fanns i följande loopar:

// Vänta tills tråden som kör 'fractal_print' har skrivit ut
// raden, så att vi kan börja med nästa.
while (d->thread_start[my_id] == false)    // <-- Busy wait
    ;
d->thread_start[my_id] = false;


// Vänta på att alla trådar är klara.
while (data.threads_done < THREAD_COUNT)   // <-- Busy wait
    ;
data.threads_done = 0;


c:

Se koden nedan.

d:

Se koden nedan.

 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"

const char ALPHABET[] = " .:-=+*#%@";
const int ALPHABET_COUNT = strlen(ALPHABET);
const int MAX_ITERATIONS = 100;

/**
 * En punkt.
 */
struct point {
    double x;
    double y;
};

// Beräkna avståndet från punkten (0, 0) till den givna punkten.
double point_length(struct point p) {
    return sqrt(p.x * p.x + p.y * p.y);
}

// En funktion som genererar en fraktal. Funktionen tar en struct
// point som parameter och returnerar en struct point.
typedef struct point (*fractal_generator)(struct point);


/**
 * Beskrivning av en fraktal.
 */
struct fractal {
    // Funktion som genererar fraktalen.
    fractal_generator generator;

    // Zoom-nivå. Större tal ger mer inzoomad bild.
    double zoom;

    // Position i fraktalen som är centrerad i bilden.
    struct point center;

    // Höjd och bredd av bilden som ska ritas ut.
    int height;
    int width;
};


// Beräkna värdet i en punkt i fraktalen. Värdet motsvarar antalet
// iterationer som krävs innan punkten som produceras av generatorn är
// för stor. Detta värde översätts sedan till en "färg" av funktionen
// 'fractal_iter_to_char'.
// Du kan anta att funktionen är intern, och bara anropas från andra
// funktioner som hör till datastrukturen 'fractal'.
static int fractal_compute_point(struct fractal *f, struct point p) {
    struct point current = { 0, 0 };
    fractal_generator generator = f->generator;

    for (int i = 0; i < MAX_ITERATIONS; i++) {
        if (point_length(current) > 1000)
            return i;

        current = (*generator)(current);

        current.x += p.x;
        current.y += p.y;
    }

    return MAX_ITERATIONS - 1;
}

// Översätt ett antal iterationer till en "färg" (dvs. ett tecken).
// Du kan anta att funktionen är intern, och bara anropas från andra
// funktioner som hör till datastrukturen 'fractal'.
static char fractal_iter_to_char(int iterations) {
    int elem = iterations * ALPHABET_COUNT / MAX_ITERATIONS;
    if (elem < 0)
        elem = 0;
    if (elem >= ALPHABET_COUNT)
        elem = ALPHABET_COUNT - 1;
    return ALPHABET[elem];
}

// Antal trådar som 'fractal_print' ska använda.
#define THREAD_COUNT 4

/**
 * Data delad mellan 'fractal_print' och 'fractal_worker'.
 */
struct worker_data {
    // Fraktalen som manipuleras.
    struct fractal *fractal;

    // Buffer som innehåller en rad av fraktalen.
    char *buffer;

    // Nuvarande rad (y-koordinat) som buffern innehåller.
    int current_row;

    // Nästa kolumn (x-koordinat) som ska skapas i buffern.
    int next_col;

    // Lås för "next_col".
    struct lock col_lock; // <-- (d)

    // Räknare för att generera tråd-id:n.
    int threads_active;

    // Ska tråden med id 'x' starta med 'current_row'?
    struct semaphore thread_start[THREAD_COUNT]; // <-- (c)

    // Antal trådar som är klara med 'current_row'.
    struct semaphore threads_done; // <-- (c)
};

// Funktion som används av 'fractal_print' för att distribuera arbetet
// på flera trådar. Du kan anta att funktionen är intern, och bara
// anropas från 'fractal_print' nedan.
static void fractal_worker(void *data) {
    struct worker_data *d = data;
    struct fractal *f = d->fractal;

    double xscale = 1.0 / (f->width * f->zoom);
    double yscale = 1.0 / (f->height * f->zoom);

    double xshift = xscale * f->width / 2 - f->center.x;
    double yshift = yscale * f->height / 2 - f->center.y;

    // Generera ett tråd-id för den här tråden.
    int my_id = atomic_add(&d->threads_active, 1);

    // Behandla alla rader.
    for (int row = 0; row < f->height; row++) {
        // Vänta tills tråden som kör 'fractal_print' har skrivit ut
        // raden, så att vi kan börja med nästa.
        sema_down(&d->thread_start[my_id]); // <-- (c)

        int y = d->current_row;

        // Beräkna alla pixlar i raden tillsammans med de andra trådarna.
        while (true) {
            // Hitta en x-koordinat som är ledig.
            lock_acquire(&d->col_lock); // <-- (d)
            int x = d->next_col++;
            lock_release(&d->col_lock); // <-- (d)
            if (x >= f->width)
                break;

            // Beräkna koordinatens "färg".
            struct point p;
            p.x = x * xscale - xshift;
            p.y = y * yscale - yshift;
            int iterations = fractal_compute_point(f, p);
            char ch = fractal_iter_to_char(iterations);

            // Skriv till buffern.
            d->buffer[x] = ch;
        }

        // Meddela att vi är klara.
        sema_up(&d->threads_done); // <-- (c)
    }
}

// Skriv ut en fraktal enligt den specifikation som finns inuti 'f'.
// 'to' är en ström som beskriver vart fraktalen ska sparas. I och
// med detta så ska det gå att köra 'fractal_print' på flera trådar
// samtidigt. Exempelvis för att spara en fraktal till en fil
// och en annan till skärmen. Du kan anta att 'to' inte är delad
// med andra trådar när funktionen anropas.
void fractal_print(struct fractal f, FILE *to) {
    // Initiera data som delas med worker-trådar.
    struct worker_data data;
    data.fractal = &f;
    data.threads_active = 0;
    data.buffer = malloc(sizeof(char) * (f.width + 1));
    lock_init(&data.col_lock);        // <-- (d)
    sema_init(&data.threads_done, 0); // <-- (c)

    // Starta trådar.
    for (int i = 0; i < THREAD_COUNT; i++) {
        sema_init(&data.thread_start[i], 0); // <-- (c)
        thread_new(&fractal_worker, &data);
    }

    for (int row = 0; row < f.height; row++) {
        // Uppdatera information om vilken rad vi ska arbeta med.
        data.next_col = 0;
        data.current_row = row;

        // Meddela trådarna att vi har uppdaterat 'current_row' och
        // 'next_col', och att de kan börja arbeta.
        for (int i = 0; i < THREAD_COUNT; i++)
            sema_up(&data.thread_start[i]); // <-- (c)

        // Vänta på att alla trådar är klara.
        for (int i = 0; i < THREAD_COUNT; i++) // <-- (c)
            sema_down(&data.threads_done);     // <-- (c)

        // Se till att buffern (fortfarande) är nollterminerad och skriv ut den.
        data.buffer[f.width] = '\0';
        fprintf(to, "%s\n", data.buffer);
    }

    // Städa.
    free(data.buffer);
}


/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */


// Mandelbrot-fraktalen. Beräknar p' = p^2
struct point mandelbrot(struct point p) {
    struct point out;
    out.x = p.x * p.x - p.y * p.y;
    out.y = p.x * p.y + p.x * p.y;
    return out;
}

// Burning-ship-fraktalen. Baserad på mandelbrot, men "normaliserar"
// punkten innan varje iteration.
struct point burning_ship(struct point p) {
    p.x = fabs(p.x);
    p.y = fabs(p.y);
    return mandelbrot(p);
}

int main(void) {
    int output_width = 90;
    int output_height = 45;

    printf("Mandelbrot:\n");
    {
        struct fractal f;
        f.generator = &mandelbrot;
        f.zoom = 0.52;
        f.center.x = -0.5;
        f.center.y = 0.0;
        f.width = output_width;
        f.height = output_height;

        fractal_print(f, stdout);
    }

    printf("Burning ship:\n");
    {
        struct fractal f;
        f.generator = &burning_ship;
        f.zoom = 0.4;
        f.center.x = -0.5;
        f.center.y = -0.5;
        f.width = output_width;
        f.height = output_height;

        fractal_print(f, stdout);

        printf("Zoomed in:\n");
        f.zoom = 0.7;
        f.center.x = -0.7;
        f.center.y = -0.7;

        fractal_print(f, stdout);
    }

    return 0;
}
