/*

a:

Antag två trådar (A och B) som kör följande anrop på samma "seat":
Tråd A: seat_book(&seat, 1, 0, 1);
Tråd B: seat_book(&seat, 2, 0, 1);

Tråd A: kontrollerar att sections[0] är -1.
Tråd B: kontrollerar att sections[0] är -1.
Tråd A: tilldelar 1 till sections[0]
Tråd B: tilldelar 2 till sections[0]

Platsen är nu dubbelbokad - dvs. båda trådarna kommer att meddela att platsen är bokad.

b:

Se koden.

c:

Mutual exclusion - det finns delar i koden som endast en tråd får köra samtidigt.
No preemption of resources - trådar måste frivilligt lämna tillbaka de resurser de har allokerat.
Hold and wait - trådar håller i en resurs (ex.vis lås) medan de väntar på en annan resurs.
Circular wait - trådar väntar på varandra i en cykel (med hjälp av hold and wait)

d:

Mutual exclusion - uppfyllt, eftersom vi använder lås.
No preemption of resources - uppfyllt, låsen vi använder måste släppas av tråden själv.
Hold and wait - uppfyllt, trådar håller i en resurs medan de väntar på en annan resurs inuti seat_book (vi kan hålla mer än 1 lås)
Circular wait - ej uppfyllt, eftersom låsen alltid tas i samma ordning (från låga index till höga index)

I och med att circular wait inte är uppfyllt så kan deadlock ej uppstå.

e:

Implementationen nedan garanterar att tråd A kommer att lyckas.

Om tråd B låser element 1 först så kommer tråd A att behöva vänta tills tråd B är helt klar. Tråd B
kommer då att ha återställt innehållet i arrayen så att det innehåller -1. Detta gör att tråd A
kommer att lyckas med sin bokning. Om tråd A kommer först så kommer den helt enkelt att lyckas med
bokningen innan B får en chans att börja boka platser.

f:

Detta är inte garanterat i och med att print-funktionen endast tar ett lås åt gången. Exempelvis kan
följande ske:

- Print-funktionen börjar skriva ut fram till och med delsträcka 2.
- En bokning sker i en annan tråd, på delsträcka 1-4. Denna kan ske eftersom print-funktionen inte håller några lås.
- Print-funktionen fortsätter att skriva ut från och med delsträcka 3.

Detta ger en utskrift där endast vissa delsträckor är utskrivna.

 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"


/**
 * Datastruktur som representerar en plats på tåget.
 *
 * Du kan anta att denna datastruktur bara används från funktioner som
 * börjar med 'seat_' och 'train_'. Dvs. datastrukturen är intern.
 */
struct seat {
    // Array som beskriver vem som har bokat platsen i varje delsträcka
    // av tågets färd. Tal mindre än 0 betyder att platsen är ledig på
    // sträckan i fråga. Ett tal större än eller lika med noll betyder
    // att kunden med det kundnumret har bokat platsen.
    int *sections;

    // Lås för 'sections'.
    struct lock *locks; // <-- (b)

    // Antal delsträckor i 'sections'-arrayen.
    int section_count;
};

// Initiera en plats med ett förutbestämt antal delsträckor.
static void seat_init(struct seat *seat, int sections) {
    seat->sections = malloc(sizeof(int) * sections);
    seat->locks = malloc(sizeof(struct lock) * sections); // <-- (b)
    seat->section_count = sections;

    for (int i = 0; i < sections; i++) {
        seat->sections[i] = -1;
        lock_init(&seat->locks[i]); // <-- (b)
    }
}

// Destruera en plats. Avallokerar *inte* seat i och med att 'init'
// inte allokerar minne.
static void seat_destroy(struct seat *seat) {
    free(seat->sections);
}

// Inspektera ifall en plats är bokad eller inte. Returnerar -1 om
// platsen är obokad på den angivna delsträckan, annars returneras vem
// som har bokat platsen.
static int seat_is_occupied(struct seat *seat, int section) {
    lock_acquire(&seat->locks[section]); // <-- (b)
    int temp = seat->sections[section];
    lock_release(&seat->locks[section]); // <-- (b)
    return temp;
}

// Försök att boka en plats på en delsträcka. 'customer_id' anger vem
// som ska boka platsen, 'section_from' och 'section_to' anger vilken
// delsträcka som ska bokas. 'section_from' är alltid mindre än
// 'section_to'. Returnerar 'true' om platsen var ledig på den angivna
// delsträckan. Annars returneras 'false'.
static bool seat_book(struct seat *seat, int customer_id, int station_from, int station_to) {
    bool ok = true;
    int filled;

    // Försök boka alla delsegment.
    for (filled = station_from; filled < station_to; filled++) {
        lock_acquire(&seat->locks[filled]); // <-- (b)
        if (seat->sections[filled] < 0) {
            seat->sections[filled] = customer_id;
        } else {
            lock_release(&seat->locks[filled]); // <-- (b)
            ok = false;
            break;
        }
    }

    // Om platsen inte var ledig hela sträckan, avboka de delar vi redan har bokat.
    for (int i = station_from; i < filled; i++) {
        if (!ok) {
            seat->sections[i] = -1;
        }
        lock_release(&seat->locks[i]); // <-- (b)
    }

    return ok;
}


/**
 * Datastruktur som beskriver alla platser på ett tåg.
 *
 * Du kan anta att endast funktioner som börjar på 'train_' kan komma
 * åt medlemmarna i datastrukturen. Däremot är alla funktioner
 * 'train_' åtkomliga till alla delar av programmet.
 */
struct train {
    // Alla platser.
    struct seat *seats;

    // Antal platser.
    int seat_count;

    // Antal delsträckor som tåget ska åka.
    int section_count;
};

// Skapa en instans av datastrukturen 'train'. Tåget har 'seats'
// platser, och ska åka 'sections' delsträckor.
struct train *train_create(int seats, int sections) {
    struct train *train = malloc(sizeof(struct train));

    train->seat_count = seats;
    train->section_count = sections;
    train->seats = malloc(sizeof(struct seat) * seats);

    for (int i = 0; i < seats; i++) {
        seat_init(&train->seats[i], sections);
    }

    return train;
}

// Förstör ett tåg.
void train_free(struct train *train) {
    for (int i = 0; i < train->seat_count; i++) {
        seat_destroy(&train->seats[i]);
    }
    free(train->seats);
    free(train);
}

// Boka en plats på en delsträcka. Om en ledig plats fanns på en
// delsträcka bokas platsen, och platsens nummer returneras. Annars
// bokas ingenting, och -1 returneras.
int train_book(struct train *train, int customer_id, int station_from, int station_to) {
    for (int i = 0; i < train->seat_count; i++) {
        if (seat_book(&train->seats[i], customer_id, station_from, station_to)) {
            return i;
        }
    }
    return -1;
}

// Hjälpfunktion för att skriva ut en horisontell linje.
static void print_hline(int length) {
    for (int i = 0; i < length; i++)
        printf("-");
    printf("\n");
}

// Skriv ut en karta över vilka platser som är bokade på vilka
// delsträckor i terminalen. Du kan anta att maximalt en tråd kör
// funktionen samtidigt. Det kan dock hända att andra trådar kör
// 'train_book' samtidigt som en utskrift sker.
void train_print(struct train *train) {
    print_hline(1 + train->seat_count * 3);

    for (int seg = 0; seg < train->section_count; seg++) {
        printf("|");
        for (int i = 0; i < train->seat_count; i++) {
            int occupied = seat_is_occupied(&train->seats[i], seg);
            if (occupied >= 0)
                printf("%2d|", occupied);
            else
                printf("  |");
        }

        printf("\n");

        print_hline(1 + train->seat_count * 3);
    }
}


/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */

struct semaphore done;

void worker(void *data) {
    struct train *train = data;

    train_book(train, 10, 0, 8);
    train_book(train, 11, 4, 10);

    sema_up(&done);
}


int main(void) {
    sema_init(&done, 0);
    struct train *train = train_create(5, 10);
    thread_new(&worker, train);

    train_book(train, 1, 0, 8);
    train_book(train, 2, 0, 4);
    train_book(train, 3, 4, 9);

    sema_down(&done);
    train_print(train);

    train_free(train);

    return 0;
}
